/*
https://www.oreilly.com/library/view/learning-java-4th/9781449372477/ch08s10.html
*/

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

/**
   Implement our own "thread pool". A "thread pool" is
   two objects, an array of worker threads and an array
   of blocking queues (one queue per worker thread).

   Each blocking queue holds Runnable objects that will
   be called(!) by the corresponding worker thread. The
   key idea is that the Runnable objects are not used to
   create threads (the worker thread already exists), the
   run() method in each Runnable object is called by the
   worker thread. So the worker thread executes the Runnable's
   run method (to completion!) and then the worker thread
   gets another Runnable object, from its blocking queue, and
   calls its run() method until the worker thread gets interrupted.

   Having one blocking queue per worker thread (as opposed to a
   single blocking queue shared by all the worker threads) has an
   advantage and a disadvantage. The advantage is that the worker
   threads do not block each other when they go to their own blocking
   queue to get a Runnable object (we don't have "lock contention").
   The disadvantage is that we may have poor "load balancing". When
   a worker thread's blocking queue is empty, that worker thread
   blocks on the empty queue and ends up not doing any work for a
   while, even though there may be Runnable items available from
   other queues. Having an idle worker thread, while there is work
   that needs to be done, is what we mean by poor load balancing.
*/
public class MyThreadPool_v1
{
   private final int poolSize;
   private final Thread[] threadPool;
   private final BlockingQueue<Runnable>[] taskQueue;
   private int currentQueue = 0;

   @SuppressWarnings({"unchecked","rawtypes"})
   public MyThreadPool_v1(int poolSize)
   {
      this.poolSize = poolSize;
      threadPool = new Thread[poolSize];
      taskQueue= new BlockingQueue[poolSize]; // unavoidable warnings

      for (int i = 0; i < poolSize; ++i)
      {
         taskQueue[i] = new LinkedBlockingQueue<Runnable>();
      }

      for (int i = 0; i < poolSize; ++i)
      {
         threadPool[i] = new WorkerThread(taskQueue[i]);
         threadPool[i].start(); // start a worker, which will block on its empty queue
      }
   }

   /**
      Submit new tasks to the task queues in round robin fashion.
   */
   public void execute(Runnable task)
   {
      taskQueue[currentQueue].add(task);
      ++currentQueue;
      currentQueue = currentQueue % poolSize;
   }

   public void shutdown()
   {
      try
      {
         for (int i = 0; i < poolSize; ++i)
         {
            threadPool[i].interrupt();
         }
         for (int i = 0; i < poolSize; ++i)
         {
            threadPool[i].join();
         }
      }
      catch (InterruptedException e)
      {
      }
   }
}

/**
   This private class implements the worker threads
   that are in the "thread pool". Each worker thread
   has a queue of Runnable tasks that it should execute.
   A worker thread executes the run() method in each
   Runnable task removed from the task queue.

   Notice how the Runnable objects are NOT used to create
   threads! The Runnable objects are simply convenient
   containers for the code that makes up a task.
*/
class WorkerThread extends Thread
{
   private final BlockingQueue<Runnable> taskQueue;

   public WorkerThread(BlockingQueue<Runnable> taskQueue)
   {
      this.taskQueue = taskQueue;
   }

   @Override
   public void run()
   {
      boolean done = false;
      while ( (! done) || (! taskQueue.isEmpty()) )
      {
         try
         {
            Runnable task = taskQueue.take();
            // Execute the Runnable's run() method in this thread.
            task.run();

            if ( Thread.interrupted() )
               done = true;
         }
         catch (InterruptedException e)
         {
            done = true;
         }
      }
      System.out.println("***** Worker thread is done.");
   }
}
