/*
   https://docs.oracle.com/en/java/javase/16/docs/api/java.base/java/math/BigInteger.html
   https://eli.thegreenplace.net/2013/01/16/python-paralellizing-cpu-bound-tasks-with-concurrent-futures
*/

import java.math.BigInteger;
import java.util.List;
import java.util.ArrayList;

/**
   This is a brute force, slow way to determine
   if an integer is a prome. It makes for a good
   computational task to experiment with parallelism.
*/
public class Primes
{
   public static boolean isPrime(long n)
   {
      if (n % 2 == 0)
      {
         return n == 2 ? true : false;
      }

      for (long r = 3; r * r <= n; r += 2)
      {
         //if ( isPrime(r) && (n % r == 0) )
         if ( n % r == 0 )
         {
            return false;
         }
      }
      return true;
   }


   public static boolean isPrime(BigInteger n)
   {
      if  (n.divide(BigInteger.valueOf(2)).equals(BigInteger.ZERO) )
      {
         return n == BigInteger.valueOf(2) ? true : false;
      }

      for (BigInteger r = new BigInteger("3"); r.pow(2).compareTo(n) <= 0; r = r.add(BigInteger.valueOf(2)))
      {
         if ( n.remainder(r).equals(BigInteger.ZERO) )
         //if ( isPrime(r) && n.remainder(r).equals(BigInteger.ZERO) )
         {
            return false;
         }
      }
      return true;
   }



   /**
      A naive factorization method. Take integer n, return list of factors.
   */
   public static List<Long> factorize(long n)
   {
      List<Long> factors = new ArrayList<>();

      factors.add(n);
      factors.add(1L);

      final long startTime = System.currentTimeMillis();

      if (n < 2)
         return factors;

      long p = 2;

      boolean done = false;
      while (! done)
      {
         if (n == 1)
            done = true;

         long r = n % p;
         if (r == 0)
         {
            factors.add(p);
            n = n / p;
         }
         else if (p * p >= n)
         {
            factors.add(n);
            done = true;
         }
         else if (p > 2)
         {
            p += 2; // advance in steps of 2 over odd numbers
         }
         else
         {
            p = 3; // if p == 2, get to 3
         }
      }
      factors.add(0L);
      final long stopTime = System.currentTimeMillis();
      factors.add(stopTime - startTime);
      return factors;
   }


   public static List<BigInteger> factorize(BigInteger n)
   {
      List<BigInteger> factors = new ArrayList<>();

      factors.add(n);

      final long startTime = System.currentTimeMillis();

      if (n.compareTo(BigInteger.valueOf(2)) < 0)
         return factors;

      BigInteger p = BigInteger.valueOf(2);

      boolean done = false;
      while (! done)
      {
         if ( n.equals(BigInteger.ONE) )
            done = true;

         BigInteger r = n.remainder(p);
         if ( r.equals(BigInteger.ZERO) )
         {
            factors.add(p);
            n = n.divide(p);
         }
         else if (p.pow(2).compareTo(n) >= 0)
         {
            factors.add(n);
            done = true;
         }
         else if (p.compareTo(BigInteger.valueOf(2)) > 0)
         {
            p = p.add(BigInteger.valueOf(2)); // advance in steps of 2 over odd numbers
         }
         else
         {
            p = BigInteger.valueOf(3); // if p == 2, get to 3
         }
      }
      factors.add(BigInteger.ZERO);
      final long stopTime = System.currentTimeMillis();
      factors.add(BigInteger.valueOf(stopTime - startTime));
      return factors;
   }
}
